### MP Fénelon Sainte-Marie --- Informatique
### Chapitre 4 --- Complexité

##############################################################################
print("-------- EX1 --- Du danger d'une programmation insouciante --------")

import time

def u1(n):
    if n==0:
        return 2
    else:
        return (u1(n-1) + (3/u1(n-1))) / 2

def u2(n):
    if n==0:
        return 2
    else:
        x = u2(n-1)
        return (x + (3/x)) / 2

N = 20
t = time.time()
for n in range(N):
    u1(n)
print("Pour N=",N," algo naïf en",time.time()-t)
t = time.time()
for n in range(N):
    u2(n)
print("Pour N=",N," algo rusé en",time.time()-t)

t = time.time()
resu = u2(900)
print("Avec algo rusé u(900)=",resu," se calcule en",time.time()-t)

##############################################################################
print("---------------- EX2 --- Algorithme de Horner ----------------")

def Horner(L,x):
    n=len(L)
    resu=0
    nb=0
    for i in range(0,n):
        resu=L[n-1-i]+x*resu
        nb=nb+2
    print("nombre d'opérations ",nb)
    return resu

def HornerR(L,x):
    if L==[]:
        return 0
    else:
        a = L.pop(0)
        return a + x * HornerR(L,x)

L=[1,2,1] # polynôme 1+2x^2
resu=Horner(L,7)
print("P(x)=",resu)
resu=HornerR(L,7)
print("P(x)=",resu)

# Pour évaluer un polynôme de degré n, à n+1 coefficents, l'algorithme de
# Horner nécessite 2*(n+1) opérations : n+1 additions et n+1 multiplications.

# A comparer avec la méthode traditionnelle : P(x)=sum_{k=0}^n a_kx^k.
# Le calcul de a_kx^k nécessite k multiplications.
# Total : n additions et n*(n+1)/2 multiplications.
# Complexité équivalente à n^2/2.

"""

##############################################################################
print("---------------- EX3 --- Déterminant ----------------")

# Pour chaque permutation sigma :
# - 1 appel à epsilon(sigma) nécessite n*(n-1)/2 comparaisons ;
# - n multiplications.
# Total : n! * (n+O(n^2)) opérations soit O( (n+2)! ) opérations.

##############################################################################
print("---------------- EX4 --- Tri sélection ----------------")

def TriSelection(L):
    n=len(L)
    for i in range(0,n-1):
        j=i # indice du minimum
        for k in range(i+1,n-1):
            if L[k]<L[j]:  # si on trouve un plus petit
                j=k        # alors on garde en mémoire son indice
        temporaire=L[j]    # et uniquement à la fin on permute L[i] et L[j]
        L[j]=L[i]
        L[i]=temporaire
        #print("terme ",i," à sa place :",L)
    return L

Liste=[19,9,17,1,2,10]
print("liste à trier :",Liste)
print("Liste triée   :",TriSelection(Liste))

# Nombre de comparaisons : (n-1)+(n-2)+...+1=n*(n-1)/2 = Theta(n^2)
# Nombre d'échanges, dans le pire des cas : n.

##############################################################################
print("---------------- EX5 --- Tri insertion ----------------")

def TriInsertion(L):
    n=len(L)
    for i in range(1,n):
        #print("etape i=",i,"on insère ",L[i])
        a=L[i]
        j=i-1
        while L[j]>a and j>-1:
            L[j+1]=L[j]
            j=j-1
        L[j+1]=a
        #print("cela donne",L)
    return L

Liste=[19,9,17,1,2,10]
print("liste à trier :",Liste)
print("Liste triée   :",TriInsertion(Liste))

# Nombres de comparaisons à l'étape i : au plus i.
# Nombre total de comparaisons, dans le pire des cas :
#     somme(i,i=2..n) = (n+2)*(n-1)/2 = Theta(n^2)
# Nombre d'échanges, dans le pire des cas (quand liste triée à l'envers): 
#     1+2+...+(n-1) = n*(n-1)/2 = Theta(n^2)

##############################################################################
print("---------------- EX6 --- Exponentiation rapide ----------------")

def expo(a,n):
    resu=a
    for i in range(0,n-1):
        resu=resu*a
    return resu

import time
A=101024
N=50602
A=100000
N=50000

def fexpo(a,n):
    global nb_ops
    global appels
    if n==0: return 1
    else:
        q=n//2
        r=n%2
        prod=fexpo(a,q)
        appels=appels+1
        prod=prod*prod
        nb_ops=nb_ops+1
        if r==1:
            prod=prod*a
            nb_ops=nb_ops+1
        return prod

#print("pour a=",A," et n=",N," a^n=",fexpo(A,N)," par fexpo")
temps=time.time()
print("pour a=",A," et n=",N," a^n  par le calcul rusé")
nb_ops=0
appels=0
resu2=fexpo(A,N)
print("calcul effectué en ",time.time()-temps,"secondes")

temps=time.time()
print("pour a=",A," et n=",N," a^n  par le calcul classique")
resu1=expo(A,N)
print("calcul effectué en ",time.time()-temps,"secondes")

print("nb d'opérations pour fexpo :",nb_ops)
print("nb d'appels récursifs de fexpo :",appels)
#import math
#print(math.pow(A,N))

# Pour $k=0$, on a $n=1$ et une seule itération suffira.
# Supposons que $(H_k)$ est vraie.
# Soit maintenant un entier $n$ tel que $2^{k+1}\le n<2^{k+2}$.
# Par division euclidienne de $n$ par $2$, il vient $n=2q+r$ avec $r\in\{0,1\}$.
# Il vient alors $2^k\le q<2^{k+1}$.

# \`A la première itération, la variable $n$ prend la valeur $q$.
# D'après $(H_k)$, il reste $k+1$ itérations à effectuer pour calculer $a^q$.
# On aura donc effectué $k+2$ itérations au total.
# D'où le résultat par récurrence.

# L'inégalité $2^k\le n<2^{k+1}$ donne $k\ln2\le\ln(n)<(k+1)\ln2$
# puis $k\ln2\sim\ln n$.

##############################################################################
print("---------------- EX7 --- Matrices creuses ----------------")

import numpy as np
import numpy.random as rd

n = 1000

pourcentage = 0.9 # pourcentage de coefficients nuls
A = np.zeros((n,n))
L = []
for i in range(n):
    for j in range(n):
        if rd.random()>pourcentage:
            nombre = rd.randint(n)
            A[i,j] = nombre
            L.append([i,j,nombre])

def SommeNaive(M,n):
    s = 0
    for i in range(n):
        for j in range(n):
            s = s + M[i,j]
    return s

def SommeEvoluee(T):
    m = len(T)
    s = 0
    for i in range(m):
        s = s + T[i][2]
    return s

print("matrice aléatoire de taille",n,"avec",100*pourcentage,"% de coeffs nuls")
t = time.time()
print("somme naive",SommeNaive(A,n)," en ",time.time()-t,"secondes")

t = time.time()
print("somme évoluée",SommeEvoluee(L)," en ",time.time()-t,"secondes")
print("nombre de calculs : ",len(L),"au lieu de",n*n)

"""